import os
import logging

from urllib.error import URLError
from torchvision.datasets.utils import download_and_extract_archive, download_url, check_integrity


class DatasetDownloader():
    # responsible for downloading a prespecified set of images/catalogs to a directory
    # supports GalaxyDataset via composition

    def __init__(self, root, resources, images_to_spotcheck=None, image_dirname='images'):
        self.root = root
        self.image_dir = os.path.join(self.root, image_dirname)
        self.resources = resources
        self.images_to_spotcheck = images_to_spotcheck

    def download(self) -> None:
        """Download the data if it doesn't exist already."""

        if self._check_exists():
            return

        os.makedirs(self.root, exist_ok=True)

        # download files
        for url, md5 in self.resources:
            filename = os.path.basename(url)
            try:
                logging.info(f"Downloading {url}")
                if url.endswith('.tar.gz') or url.endswith('.zip'):
                    download_and_extract_archive(
                        url, download_root=self.root, filename=filename, md5=md5)
                else:  # don't try to extract archive, just download
                    download_url(url, root=self.root, filename=filename, md5=md5)
            except URLError as error:
                logging.info(f"Failed to download (trying next):\n{error}")
                continue


    def _check_exists(self) -> bool:
        # takes a few seconds for the image .zip
        logging.info('Checking integrity of resources')
        resources_downloaded = all([
            check_integrity(
                os.path.join(self.root, os.path.basename(res)),
                md5
            )
            for res, md5 in self.resources])
        logging.info('Resources downloaded: {}'.format(resources_downloaded))

        images_unpacked = all([
            os.path.isfile(os.path.join(self.image_dir, image_loc)) for image_loc in self.images_to_spotcheck
        ])

        logging.info('Images unpacked: {} ({}, {})'.format(images_unpacked, self.image_dir, self.images_to_spotcheck))

        return resources_downloaded & images_unpacked






import pandas as pd
import numpy as np
import torch
from torch.utils.data import Dataset
from PIL import Image
import simplejpeg


import torch
import os
import logging

import numpy as np
import pandas as pd


# https://pytorch.org/tutorials/beginner/basics/data_tutorial.html
class GalaxyDataset(Dataset):
    def __init__(self, catalog: pd.DataFrame, label_cols, transform=None, target_transform=None):
        # downloaded dataset where everything is all prepared, e.g. from prepared_datasets.gz2.gz2_setup()
        # reads a catalog of jpeg images, assumed of galaxies
        # if transform is from albumentations, datamodule should know about the transform including 
        # catalog should be split already
        # should have correct image locations under file_loc

        # internal catalog has id_str as index for query speed
        # (dataset.catalog still returns the int-indexed catalog via the catalog method below)
        # self._catalog = catalog.copy().set_index('id_str', verify_integrity=True)
        self.catalog = catalog
        
        self.label_cols = label_cols
        self.transform = transform
        self.target_transform = target_transform

    def __len__(self):
        return len(self.catalog)

    # @property
    # def catalog(self):
    #     return self._catalog.reset_index(drop=False)

    def __getitem__(self, idx):
        #the index is id_str so can use that for quick search on 1M+ catalo
        # galaxy = self._catalog.loc[idx]
        galaxy = self.catalog.iloc[idx]
        # option A
        # img_path = galaxy['file_loc']
        # image = read_image(img_path) # PIL under the hood: Returns CHW Tensor.
        # option B - tiny bit faster when CPU-limited
        with open(galaxy['file_loc'], 'rb') as f:
            # image = torch.from_numpy(decode_jpeg(f.read()).transpose(2, 0, 1))  # CHW tensor
            try:
                image = Image.fromarray(decode_jpeg(f.read()))  # HWC PIL image via simplejpeg
            except Exception as e:
                logging.critical('Cannot load {}'.format(galaxy['file_loc']))
                raise e
        label = get_galaxy_label(galaxy, self.label_cols)

        # logging.info((image.shape, torch.max(image), image.dtype, label))  # always 0-255 uint8

        if self.transform:
            # a CHW tensor, which torchvision wants. May change to PIL image.
            image = self.transform(image)

        if self.target_transform:
            label = self.target_transform(label)

        # logging.info((image.shape, torch.max(image), image.dtype, label))  #  should be 0-1 float
        return image, label


def load_encoded_jpeg(loc):
    with open(loc, "rb") as f:
        return f.read()  # bytes, not yet decoded


def decode_jpeg(encoded_bytes):
    return simplejpeg.decode_jpeg(encoded_bytes, fastdct=True, fastupsample=True)


def get_galaxy_label(galaxy, label_cols):
    # no longer casts to int64, user now responsible in df. If dtype is mixed, will try to infer with infer_objects
    return galaxy[label_cols].infer_objects().values.squeeze()  # squeeze for if there's one label_col




# # https://pytorch.org/vision/stable/_modules/torchvision/datasets/mnist.html for download process inspiration
# class GalaxyDatasetFactory():

#     def __init__(self, root, catalog=None, label_cols=None,  transform=None, target_transform=None) -> None:

#         # can use target_transform to turn counts into regression or even classification
#         # will need another step to drop rows, in DataModule probably

#         self.root = root

#         catalog, label_cols = self.adjust_catalog_and_labels(root, catalog, label_cols)

#         super().__init__(catalog=catalog, label_cols=label_cols,
#                          transform=transform, target_transform=target_transform)


#     def adjust_catalog_and_labels(self, root, catalog, label_cols):
#         if not self._check_exists():
#             raise RuntimeError(
#                 "Dataset not found. You can use download=True to download it")

#         if catalog is None:
#             logging.info('Loading dataset with default (unsplit) catalog')
#             catalog = pd.read_parquet(os.path.join(root, self.default_catalog_loc))
#             catalog['file_loc'] = catalog['filename'].apply(
#                 lambda x: os.path.join(self.image_dir, x))
#         else:
#             logging.info(
#                 'Overriding default catalog with user-provided catalog (length {})'.format(len(catalog)))
#             assert isinstance(catalog, pd.DataFrame)
#             # will always check label_cols as well, below
#             assert 'file_loc' in catalog.columns.values

#         if label_cols is None:
#             logging.info('Loading dataset with default label columns')
#             label_cols = self.default_label_cols
#         else:
#             logging.info('User provided GZ2 dataset with custom label cols {}'.format(label_cols))
#                 # label_cols_present = [col in catalog.columns.values for col in label_cols]
#             missing_label_cols = set(label_cols) - set(catalog.columns.values)
#             if len(missing_label_cols) > 0:
#                 raise KeyError(f'User asked for label columns not present in catalog:\n{missing_label_cols} not in \n{catalog.columns.values}')
#         return catalog,label_cols



import os

import pandas as pd
from pytorch_galaxy_datasets import galaxy_datamodule, galaxy_dataset, download_utils

# TODO could eventually refactor this out of Zoobot as well
from zoobot.shared import label_metadata

# DR8 will be basically the same


class DecalsDR5Dataset(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

        catalog, label_cols = decals_dr5_setup(root, train, download)

        super().__init__(catalog, label_cols, transform, target_transform)


def decals_dr5_setup(root, train, download):
    resources = [
        ('https://dl.dropboxusercontent.com/s/1tuehajonhgv8a2/decals_dr5_ortho_train_catalog.parquet', 'a0cd74edc073fdff068370f6eefeb802'),  # train catalog
        ('https://dl.dropboxusercontent.com/s/3vo6hjlnbqgzuxz/decals_dr5_ortho_test_catalog.parquet', '55820e3712b22e587f6971e4b6c73dfe'),  # test catalog
        ('https://dl.dropboxusercontent.com/s/bs6jp0mkgiekhww/decals_dr5_images.tar.gz', '1347de4c8df4ec579d5a58241c1f280b')  # the images
    ]
    images_to_spotcheck = ['J073/J073013.60+242930.0.jpg']

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    

    # useful_columns = label_cols + ['subfolder', 'filename']
    if train:
        train_catalog_loc = os.path.join(root, 'decals_dr5_ortho_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc)
    else:
        test_catalog_loc = os.path.join(root, 'decals_dr5_ortho_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(root, downloader.image_dir, x['subfolder'], x['filename']), axis=1)

    # default, but not actually used here
    label_cols = label_metadata.decals_dr5_ortho_label_cols
    return catalog, label_cols



# def _temp_adjust_catalog_dtypes(catalog, label_cols):
#     # enforce datatypes
#     for answer_col in label_cols:
#         catalog[answer_col] = catalog[answer_col].astype(int)
#     return catalog


if __name__ == '__main__':

    # first download is basically just a convenient way to get the images and canonical catalogs
    dr5_dataset = DecalsDR5Dataset(
        root='/nvme1/scratch/walml/repos/pytorch-galaxy-datasets/roots/decals_dr5',
        train=True,
        download=False
    )
    dr5_catalog = dr5_dataset.catalog
    adjusted_catalog = dr5_catalog.sample(1000)

    # user will probably tweak and use images/catalog directly for generic galaxy catalog datamodule
    # (which makes its own generic datasets internally)
    datamodule = galaxy_datamodule.GalaxyDataModule(
        label_cols=label_metadata.decals_dr5_ortho_label_cols,
        catalog=adjusted_catalog
    )

    datamodule.prepare_data()
    datamodule.setup()
    for images, labels in datamodule.train_dataloader():
        print(images.shape, labels.shape)
        break
        

# class DECaLSDR5DataModule(galaxy_datamodule.GalaxyDataModule):

#     def __init__(self, *args, **kwargs) -> None:
#         """
#         Currently identical to GalaxyDataModule - see that description
#         """
#         super().__init__(*args, dataset_class=DECaLSDR5Dataset, **kwargs)

#     def prepare_data(self):
#         DECaLSDR5Dataset(self.root, download=True)



import os

import pandas as pd
from pytorch_galaxy_datasets import galaxy_datamodule, galaxy_dataset, download_utils

# TODO could eventually refactor this out of Zoobot as well
from zoobot.shared import label_metadata


# TODO add train flag? currently full dataset
class GZ2Dataset(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

        catalog, label_cols = gz2_setup(root, train, download)  # no train arg

        super().__init__(catalog, label_cols, transform, target_transform)


def gz2_setup(root, train, download):
    resources = [
        ('https://dl.dropboxusercontent.com/s/vu77e3sh2s5c250/gz2_train_catalog.parquet', 'f489c9ec7dcf8d99f728bd00ee00b1d0'),  # the train catalog
        ('https://dl.dropboxusercontent.com/s/t8eh6f3oupndpl3/gz2_test_catalog.parquet', '8b2d74c93d35f86cc34f1d058b3b220b'),  # the test catalog
        ('https://zenodo.org/record/3565489/files/images_gz2.zip', 'bc647032d31e50c798770cf4430525c7')  # the images
    ]
    images_to_spotcheck = ['100097.jpg']

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()


    # useful_columns = label_cols + ['filename']
    useful_columns = None  # quite a variety of possiblities - load all of them
    if train:
        train_catalog_loc = os.path.join(root, 'gz2_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'gz2_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['file_loc'] = catalog['filename'].apply(
        lambda x: os.path.join(downloader.image_dir, x))

    label_cols = label_metadata.gz2_ortho_label_cols  # default, but you can ignore
    return catalog, label_cols


if __name__ == '__main__':

    # can use this directly, e.g. to visualise the dataset
    gz2_dataset = GZ2Dataset(
        root='/nvme1/scratch/walml/repos/pytorch-galaxy-datasets/roots/gz2',
        download=False
    )
    gz2_catalog = gz2_dataset.catalog  # and can get the catalog/cols for tweaking
    gz2_label_cols = gz2_dataset.label_cols

    # or, can use the setup method directly
    catalog, label_cols = gz2_setup(
        root='/nvme1/scratch/walml/repos/pytorch-galaxy-datasets/roots/gz2',
        download=False
    )
    adjusted_catalog = gz2_catalog.sample(1000)

    datamodule = galaxy_datamodule.GalaxyDataModule(
        label_cols=['smooth-or-featured_smooth'],
        catalog=adjusted_catalog
    )

    datamodule.prepare_data()
    datamodule.setup()
    for images, labels in datamodule.train_dataloader():
        print(images.shape, labels.shape)
        break


import os

import pandas as pd
from pytorch_galaxy_datasets import galaxy_datamodule, galaxy_dataset, download_utils

# TODO could eventually refactor this out of Zoobot as well
from zoobot.shared import label_metadata


class TidalDataset(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None, label_mode='coarse'):

        catalog, label_cols = tidal_setup(root, train, download, label_mode=label_mode)

        super().__init__(catalog, label_cols, transform, target_transform)


def tidal_setup(root, train, download, label_mode='coarse'):
    resources = [
        ('https://dl.dropboxusercontent.com/s/jq6dyc8q87h92qq/tidal_train_catalog.parquet', '39def8527823f6d4f332fbc209b15a32'),  # train catalog
        ('https://dl.dropboxusercontent.com/s/h8m7j7o3o11mt05/tidal_test_catalog.parquet', '12a8c6a23ddc8e7ad3f5bd597931bbe3'),  # test catalog
        ('https://dl.dropboxusercontent.com/s/icb6kgixe3imf4a/tidal_images.tar.gz', '1e10cd3a2a4a4c5fe39e447b2257e5f0')  # the images
    ]
    images_to_spotcheck = ['W1-2_color.jpg', 'W3-82_color.jpg', 'W3-852_color.jpg']

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    if label_mode == 'coarse':
        label_cols = coarse_tidal_label_cols
    elif label_mode == 'finegrained':
        label_cols = finegrained_tidal_label_cols

    useful_columns = label_cols + ['filename']
    if train:
        train_catalog_loc = os.path.join(root, 'tidal_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'tidal_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['id_str'] = catalog['filename'] # usually I have included these in the catalog itself TODO
    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(root, downloader.image_dir, x['filename']), axis=1)


    return catalog, label_cols

# defined outside for use elsewhere
coarse_tidal_label_cols = ['coarse_tidal_label']
finegrained_tidal_label_cols = ['finegrained_tidal_label']

if __name__ == '__main__':

    # first download is basically just a convenient way to get the images and canonical catalogs
    tidal_catalog, label_cols = tidal_setup(
        root='/nvme1/scratch/walml/repos/pytorch-galaxy-datasets/roots/tidal',
        train=True,
        download=False,
    )
    adjusted_catalog = tidal_catalog.sample(1000)

    # user will probably tweak and use images/catalog directly for generic galaxy catalog datamodule
    # (which makes its own generic datasets internally)
    datamodule = galaxy_datamodule.GalaxyDataModule(
        label_cols=label_cols,
        catalog=adjusted_catalog
    )

    datamodule.prepare_data()
    datamodule.setup()
    for images, labels in datamodule.train_dataloader():
        print(images.shape, labels.shape)
        break
        


import os

import pandas as pd
from pytorch_galaxy_datasets import galaxy_datamodule, galaxy_dataset, download_utils

from pytorch_galaxy_datasets.prepared_datasets import internal_urls

# TODO could eventually refactor this out of Zoobot as well
from zoobot.shared import label_metadata


class RingsDataset(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

        catalog, label_cols = rings_setup(root, train, download)

        super().__init__(catalog, label_cols, transform, target_transform)


def rings_setup(root, train, download):
    resources = [
        (internal_urls.rings_train_catalog, 'e2fb6b2bca45cd7f1c58f5b4089a5976'),  # train catalog
        (internal_urls.rings_test_catalog, '6e3f362a6e19ecd02675eaa48f6727f0'),  # test catalog
        (internal_urls.rings_images, 'd0950250436a05ce88de747e6af825b6')  # the images
    ]
    images_to_spotcheck = []

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    # useful_columns = label_cols + ['subfolder', 'filename']
    useful_columns = None
    if train:
        train_catalog_loc = os.path.join(root, 'rings_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'rings_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(root, downloader.image_dir, x['subfolder'], x['filename']), axis=1)

    label_cols = label_metadata.rings_label_cols

    return catalog, label_cols


if __name__ == '__main__':

    # first download is basically just a convenient way to get the images and canonical catalogs
    rings_datset = RingsDataset(
        root='/nvme1/scratch/walml/repos/pytorch-galaxy-datasets/roots/decals_dr5',
        train=True,
        download=False
    )
    rings_catalog = rings_datset.catalog
    adjusted_catalog = rings_catalog.sample(1000)

    # user will probably tweak and use images/catalog directly for generic galaxy catalog datamodule
    # (which makes its own generic datasets internally)
    datamodule = galaxy_datamodule.GalaxyDataModule(
        label_cols=['ring_fraction'],  # counts and totals also available
        catalog=adjusted_catalog
    )

    datamodule.prepare_data()
    datamodule.setup()
    for images, labels in datamodule.train_dataloader():
        print(images.shape, labels.shape)
        break
        


rings_train_catalog = 'https://dl.dropboxusercontent.com/s/bxa7nti8gq8bjd7/rings_train_catalog.parquet'
rings_test_catalog = 'https://dl.dropboxusercontent.com/s/76bc9mb9q1a1pft/rings_test_catalog.parquet'
rings_images = 'https://dl.dropboxusercontent.com/s/4ax8h8yjq3nbxya/rings_images.tar.gz'

legs_train_catalog = 'https://dl.dropboxusercontent.com/s/6dxquekzju7xfyq/legs_all_campaigns_ortho_dr8_only_train_catalog.parquet'
legs_test_catalog = 'https://dl.dropboxusercontent.com/s/lrqzmaishpgpnpk/legs_all_campaigns_ortho_dr8_only_test_catalog.parquet'
legs_unlabelled_catalog = 'https://dl.dropboxusercontent.com/s/uipqfl0ucovy83v/legs_all_campaigns_ortho_dr8_only_unlabelled_catalog.parquet'

legs_chunk_00 = 'https://dl.dropboxusercontent.com/s/uo8lkxjy3t8zugj/low_z_jpg_chunk_00_archive.tar.gz'
legs_chunk_01 = 'https://dl.dropboxusercontent.com/s/i9x3wtw9h7qy7jx/low_z_jpg_chunk_01_archive.tar.gz'
legs_chunk_02 = 'https://dl.dropboxusercontent.com/s/y8xibdnonp76az9/low_z_jpg_chunk_02_archive.tar.gz'
legs_chunk_03 = 'https://dl.dropboxusercontent.com/s/j8yuooyt52plrld/low_z_jpg_chunk_03_archive.tar.gz'
legs_chunk_04 = 'https://dl.dropboxusercontent.com/s/yrgthkvtc3kerxc/low_z_jpg_chunk_04_archive.tar.gz'
legs_chunk_05 = 'https://dl.dropboxusercontent.com/s/q9ukrsypfodmwyf/low_z_jpg_chunk_05_archive.tar.gz'
legs_chunk_06 = 'https://dl.dropboxusercontent.com/s/vr9v69hmgsh7ios/low_z_jpg_chunk_06_archive.tar.gz'
legs_chunk_07 = 'https://dl.dropboxusercontent.com/s/sp8jlfm41eup1bo/low_z_jpg_chunk_07_archive.tar.gz'


from pytorch_galaxy_datasets.prepared_datasets.gz2 import GZ2Dataset, gz2_setup
from pytorch_galaxy_datasets.prepared_datasets.dr5 import DecalsDR5Dataset, decals_dr5_setup

import os
# i.e. if internal_urls.py is available in the folder of this script
if os.path.isfile(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'internal_urls.py')):  
    # import the internal datasets
    from pytorch_galaxy_datasets.prepared_datasets.legs import LegsDataset, legs_setup
    from pytorch_galaxy_datasets.prepared_datasets.rings import RingsDataset, rings_setup

# allows for imports like
# from pytorch_galaxy_datasets.prepared_datasets import GZ2Dataset


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytorch_galaxy_datasets",
    version="0.0.1",
    author="Mike Walmsley",
    author_email="walmsleymk1@gmail.com",
    description="Galaxy Zoo datasets for PyTorch/Lightning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mwalmsley/pytorch-galaxy-datasets",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: GPU :: NVIDIA CUDA"
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.7"
)
